"""
Implement Q-learning with epsilon-greedy in the tabular case
"""
import numpy as np
from matplotlib import pyplot as plt
from cartpole_cs292 import CartPoleEnv


def policy_naive(obs):
    """
    A naive policy that takes Action 0 if the angle of the rod is negative and Action 1 otherwise

    :param obs: observation of the environment (state)
    :return: action value: int
    """
    ############################
    # YOUR IMPLEMENTATION HERE #

    ############################
    pass


def policy_evaluation(policy, num_episode=1000, episode_length=200):
    """
    On-policy evaluation of the policy

    :param policy: policy to be evaluated
    :param num_episode: number of episodes
    :param episode_length: max horizon of the episode (max steps)
    :return: total_reward_list: np.array(num_episode)
        total reward for each episode
    """
    total_reward_list = np.zeros(num_episode)
    for epoch in range(num_episode):
        observation = env.reset()
        total_reward = 0
        for iter in range(episode_length):
            ############################
            # YOUR IMPLEMENTATION HERE #

            ############################
            pass
        total_reward_list[epoch] = total_reward
        env.close()
    return total_reward_list


def discretize_state(obs):
    """
    Discretize the observed continuous states into a discrete states
    position [-2.4, 2.4]; velocity [-5, 5];
    pole angle [-0.418, 0.418]; angular velocity [-1, 1]

    :param obs: observation of the environment (state)
    :return: state value: int
    """
    obs_clipped = np.minimum(clip_range - 1e-10, np.maximum(-clip_range, obs))
    discrete_state = ((obs_clipped + clip_range) / step_size).astype(int)
    return np.ravel_multi_index(discrete_state, num_steps)


def policy(obs):
    """
    Select the greedy action according to a given Q-value function

    :param obs: observation
    :return: action value: int
    """
    ############################
    # YOUR IMPLEMENTATION HERE #

    ############################
    pass


def Update_Q(obs, a, r, obs_prime, lr=0.2, gamma=0.99):
    """
    Update the Q value matrix

    :param obs: observation
    :param a: action
    :param r: reward
    :param obs_prime: next state (observation)
    :param lr: learning rate
    :param gamma: discounted value
    """
    ############################
    # YOUR IMPLEMENTATION HERE #

    ############################
    pass


def q_learning_tabular():
    """
    Implement Q-learning with epsilon-greedy

    :return: None
    """
    # default parameters
    num_episode = 400
    episode_length = 200
    learning_rate = 0.1
    gamma = 1.0
    epsilon = 1
    epsilon_decay = 0.995

    expected_reward_list = []
    for epoch in range(num_episode):
        total_reward = 0
        observation = env.reset()
        for iter in range(episode_length):
            ############################
            # YOUR IMPLEMENTATION HERE #

            ############################
            pass

        print('Episode ', epoch, ', Total reward =', total_reward, ', epsilon =', epsilon)
        epsilon = epsilon * epsilon_decay
        env.close()

        if epoch % 50 == 0:
            reward_list = policy_evaluation(policy, num_episode=400)
            expected_reward_list.append(np.mean(reward_list))

    plt.plot(np.arange(1, num_episode + 1, 50), expected_reward_list)
    plt.xlabel('Episode')
    plt.ylabel('Reward')
    plt.show()
    plt.close()


if __name__ == '__main__':
    env = CartPoleEnv()

    print('Action space:', env.action_space)
    print('State space:', env.observation_space)

    # run naive policy
    naive_reward_list = policy_evaluation(policy_naive, num_episode=1000)
    print('Mean of the total reward:', np.mean(naive_reward_list))
    print('Standard deviation of the total reward:', np.std(naive_reward_list))

    # parameters for discretize the state
    clip_range = np.array([2.4, 5.0, 0.418, 1.0])
    num_steps = np.array([5, 5, 20, 5])
    step_size = clip_range * 2 / num_steps

    # tabular representation of state and Q value
    S = np.prod(num_steps)
    Q = np.zeros(shape=(S, 2))

    # run q_learning in the tabular case
    q_learning_tabular()

